//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/dollhouse/dolldrop.hpp
// -Declarations of the baseline RW Dolldrop sub-library

#pragma once

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

//////////////////
//  'Net Work :
//

// DESIGN:
// -As a simple, basic control mechanism for starters we'll use some
// protected/private classes and a few functions from here within Dolldrop.
// Later, we can add more access restrictions and some further data
// validations; improving things bit-by-bit as we go along, using just a few
// of the straightforward language features readily available to us.

// DESIGN:
// - add work-stealing queues of the Dolldrops/Mailboxes ?;
// - use S.Parent's libdispatch-based async
// - this is actually a general need, so add it to the RW Foundations instead

// DESIGN:
// void push_drop() {}
// void yank_drop() {}

// DESIGN: TODO:
// return do_atomiconn(
//   lockout();
//     connect();
//       do_req(uri);
//     disconnect();
//   unlockout();
// );

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------
/**
 */
class Dolldrop_b : public RW_slave {
 public:
  Dolldrop_b() = delete;                                  // no default ctor
  Dolldrop_b(const Hip& tag, std::string const& handle);  // tag-parm'd ctor

 protected:
  inline auto ask_drop_hdr(Hip const& tag, std::string const& uri)
      -> std::ostringstream {
    if (tag == master())                        // 1st confirm Dolldrop's Master
      return mailstop_.try_drop_hdr(t(), uri);  // then redirect to mailstop
    else
      only_master_can(master(), tag, t(), "Dolldrop_b::ask_drop_hdr()");

    ///
    return ostringstream{};  // shouldn't reach here
  }

 private:
  //----------------------------------------------------------------------------
  /**
   */
  class Mailstop {
   public:
    Mailstop(Hip const& tag)
        : owner_{tag}, dollnet_{owner_, "dolldrop_dollnet"} {}

    auto try_drop_hdr(Hip const& tag, std::string const& uri)
        -> std::ostringstream {
      return dollnet_.dl_w_header(tag, uri);  // TEMP: deliver directly for now
    }

   private:
    Hip owner_;

    class Inbox {
      //
     private:
      string data_;
    } inbox_;

    class Outbox {
      //
     private:
      string data_;
    } outbox_;

    //--------------------------------------------------------------------------
    /**
     */
    class Dollnet : public RW_slave {
     public:
      Dollnet() = delete;                                  // no default ctor
      Dollnet(Hip const& tag, std::string const& handle);  // tag-parm'd ctor

      /** Attempts an external network connection with, and download transfer
       * from, @a uri. If successful, transfer's body stream for @a uri is saved
       * out to disk file ( ./dolldrops/anons_file ).
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @return header stream data for @a uri, if present
       * @note: internally catches a specialization of @c std::runtime_error on
       * @b curl errors (but doesn't re-throw, currently)
       */
      auto dl_w_header(Hip const&         tag,  // guarded; for Master only
                       std::string const& uri) -> std::ostringstream;

    } dollnet_;

  } mailstop_{t()};  // construct this member w/ enclosing Dolldrop as owner

};  // class Dolldrop_b

//------------------------------------------------------------------------------
/**
 */
class Dolldrop : public Dolldrop_b {
 public:
  Dolldrop() = delete;                                  // no default ctor
  Dolldrop(const Hip& tag, std::string const& handle);  // tag-parm'd ctor

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return header stream data for @a uri, if present
   */
  auto pull_w_hdr(Hip const& tag, std::string const& uri)
      -> std::ostringstream {
    stamp(tag, "pull_w_hdr()");

    // TEST:
    prt_tag(tag, "asked to pull a dolldrop w/ hdr");
    cout << " - requested pull address :  '" << uri << "'\n\n";

    // DESIGN: basic validation for Type of the caller 'tag'
    //  -we can implement an 'any_of()' algorithm later if needed. but for now,
    //  Type::Bumpmaster and Type::Curator are the only valid RW types for
    //  dolldrop info requests.
    //
    if (tag.typ == Type::Bumpmaster || tag.typ == Type::Curator)
      return ask_drop_hdr(tag, uri);
    else
      error("Invalid type for dolldrops", t(), "Dolldrop::pull_w_hdr()");

    ///
    return ostringstream{};  // shouldn't reach here
  }

};  // class Dolldrop

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dollhouse_lib.v211225
// Copyright (2021)
// License MIT (Expat) https://opensource.org/licenses/MIT
