//   RW Dollnet
// ==============
// External, Inter-Networking Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/prog_bar.cpp
// -Definitions of some baseline RW Dollnet utility functions

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "curl_easy.h"

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------

constexpr int tot_ticks{60};  // how wide the progress meter bar is (in chars)
Time          start;          // will start the instant before an xfer begins

//------------------------------------------------------------------------------
/**
 * @param xfer_frac
 * @return
 */
std::string pct(float const xfer_frac) {
  auto          pct{xfer_frac * 100};
  ostringstream pctss;
  pctss << setw(6) << setfill(' ') << fixed << setprecision(1) << pct;

  return pctss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_frac
 * @param do_rev optional flag, reverse bar direction
 * @return
 */
std::string ticks(float const xfer_frac, bool const do_rev = false) {
  // part already done (the tickmarks)
  auto done = static_cast<int>(round(tot_ticks * xfer_frac));

  // part still remaining (the spaces)
  auto rem{tot_ticks - done};

  ostringstream tickss;
  if (do_rev)
    tickss << setw(rem) << setfill(' ') << "" << setw(done) << setfill('=')
           << "";
  else
    tickss << setw(done) << setfill('=') << "" << setw(rem) << setfill(' ')
           << "";

  return tickss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_tot
 * @param xfer_done
 * @return
 */
std::string m_of_n(curl_off_t const xfer_tot, curl_off_t const xfer_done) {
  ostringstream m_of_nss;
  m_of_nss << setw(6) << xfer_done << " of " << setw(6) << xfer_tot;

  return m_of_nss.str();
}

//------------------------------------------------------------------------------
/**
 * @return
 */
std::string secs() {
  Secs_f elapsed{Call_clk::now() - start};  // TODO: separate times ?
                                            // -(ie, ul vs. dl times could vary)
  ostringstream secss;
  secss << setw(4) << setfill(' ') << fixed << setprecision(1)
        << elapsed.count();

  return secss.str();
}

//------------------------------------------------------------------------------
/**
 */
enum class Dir {
  up,
  dn,
};

//------------------------------------------------------------------------------
/** The actual 'print the bar' code itself
 * @param xfer_tot
 * @param xfer_cmplt
 * @param dir
 * @param do_rev optional flag, reverse bar direction
 * @note: works for either upload or download
 */
void prt_bar(curl_off_t const xfer_tot,
             curl_off_t const xfer_cmplt,
             Dir const        dir,
             bool const       do_rev = false) {
  // cast prevents integer division
  auto xfer_frac = xfer_cmplt / static_cast<float>(xfer_tot);

  auto xfer_pct    = pct(xfer_frac);                // percent string
  auto xfer_ticks  = ticks(xfer_frac, do_rev);      // tickmarks string
  auto xfer_m_of_n = m_of_n(xfer_tot, xfer_cmplt);  // bytes string
  auto xfer_secs   = secs();                        // seconds string

  cout << ((dir == Dir::up) ? "u/l" : "d/l") << xfer_pct << "% [" << xfer_ticks
       << "]  bytes : " << xfer_m_of_n << ", " << xfer_secs << 's';
}

//------------------------------------------------------------------------------
/** Prints 'progress meter bars' for @b curl 's transfers, out to console
 *
 * This function acts as a callback, called at @b curl 's discretion while
 * transfers are in progress
 *
 * @param unused
 * @param dl_tot
 * @param dl_rcvd
 * @param ul_tot
 * @param ul_sent
 * @return default '0' value, unused
 * @note: transfer's status data are supplied by the @b curl library itself,
 * we are simply providing for the required form for it here and utilizing
 * them
 */
int prt_prog(Address,
             curl_off_t dl_tot  = 0,
             curl_off_t dl_rcvd = 0,
             curl_off_t ul_tot  = 0,
             curl_off_t ul_sent = 0) {  //

  // TEST: simple test of our 'stacked bars' display approach (using just a dl)
  // TEMP: -'upload' data is fake during this test, of course
  // ul_tot  = dl_tot;
  // ul_sent = dl_rcvd;

  // UPLOAD
  if (ul_tot > 0) {  // have an upload to display?
    prt_bar(ul_tot, ul_sent, Dir::up);

    // line ending depends if we have a simultaneous dl (stacked bars, or not)
    cout << ((dl_tot > 0) ? "\n" : "\r");
  }

  // DOWNLOAD
  if (dl_tot > 0) {  // have a download to display?
    prt_bar(dl_tot, dl_rcvd, Dir::dn, true);

    // line ending depends if we have a simultaneous ul (upline + r, or just r)
    cout << ((ul_tot > 0) ? "\u008D\r" : "\r");
  }

  // printing w/ flushing works smoother in this particular type case ('\r')
  cout << flush;

  // finally, transfer will abort w/o a '0' func ret
  return 0;
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "This apartment has super-high bandwidth illegal even to large corporations?"

// rw_dollhouse_lib.v211225
// Copyright (2021)
// License MIT (Expat) https://opensource.org/licenses/MIT
